@@ -12,18 +12,19 @@ This module provides some commonly used MSD fits. Most notably:
 + Fitting cubic splines to your MSD curve via `SplineFit`. Doing this for
   varying number ``n`` of spline points allows to discern which features of an
   empirical MSD curve are significant and which are just noise. To compare fits
-  with different number of spline points, you can use AIC; the number of free
-  parameters for each fit can be obtained from `Fit.number_of_fit_parameters
-  <bayesmsd.fit.Fit.number_of_fit_parameters>`.
+  with different number of spline points, you can use AIC.
 """
+from copy import deepcopy
+
 import numpy as np
 from scipy import optimize, interpolate
 
 import rouse
-from noctiluca.analysis.p2 import MSD
+from noctiluca.analysis import MSD
 
 from . import deco
 from .fit import Fit
+from .parameters import Parameter, Linearize
 
 class SplineFit(Fit):
     """
@@ -38,12 +39,10 @@ class SplineFit(Fit):
     from ``t in [1, T]`` to ``x in [0, 1]``:
 
     - if ``ss_order == 0``, we need ``t = inf`` to be accessible. We therefore
-      choose the compactification ``x = 4/π*arctan(log(t/T))``, such that ``t =
-      inf`` corresponds to ``x = 2``.
+      choose the compactification ``x = 4/π*arctan(log(t)/log(T))``, such that
+      ``x(t=∞) = 2`` while ``x(t=T) = 1``.
     - if ``ss_order == 1``, we simply work in log-space, and normalize: ``x =
-      log(t)/log(T)``.
-
-    These compactifications are of course applied only internally.
+      log(t)/log(T)``. Thus ``x(t=T) = 1``.
 
     For the y-coordinate of the spline we apply a simple log-transform, ``y =
     log(MSD)``. We then use the boundary condition that the second derivative
@@ -52,7 +51,7 @@ class SplineFit(Fit):
 
     Parameters
     ----------
-    data : TaggedSet of Trajectory
+    data : noctiluca.TaggedSet, pandas.DataFrame, list of numpy.ndarray
         the data to fit. See `Fit`.
     ss_order : {0, 1}
         the steady state order to assume
@@ -64,6 +63,10 @@ class SplineFit(Fit):
         model selection task over spline knots. First entry should be the
         `Fit` object used, second one should be the resulting dict with keys
         ``'params'`` and ``'logL'``.
+    motion_blur_f : float in [0, 1]
+        fractional exposure time (i.e. exposure time as fraction of the time
+        between frames). The resulting motion blur will be taken into account
+        in the fit.
 
     Notes
     -----
@@ -78,11 +81,12 @@ class SplineFit(Fit):
     transformed coordinate system (``(dt, MSD) --> (x, y) == (compactify(dt),
     log(MSD))``). Since the total extent of the data along the time axis is
     fixed, we also fix the first and last x-coordinates, such that ultimately
-    the free parameters are ``(n-2)*[x] + n*[y]``.
+    the free parameters are ``x1`` through ``x(n-2)`` and ``y0`` through
+    ``y(n-1)``
 
     See also
     --------
-    Fit, msdfit
+    Fit, bayesmsd
     """
     def __init__(self, data, ss_order, n,
                  previous_spline_fit_and_result=None,
@@ -111,13 +115,20 @@ class SplineFit(Fit):
         else: # pragma: no cover
             raise ValueError(f"Did not understand ss_order = {ss_order}")
 
-        # Parameters are (n-2)*x_spline + n*y_spline
-        # The x-coordinates of the first and last spline points are fixed
-        # For the bounds, note that x lives in the compactified interval [0, 1]
-        # (or [0, 2]), while y = log(MSD) can be any real number. We use actual
-        # 0 as bound instead of 1e-10 to signal the Profiler that this is not a
-        # logarithmic quantity.
-        self.bounds = (self.n-2)*[(self.x_first, self.x_last)] + self.n*[(-np.inf, np.inf)]
+        # x lives in the compactified interval [0, 1] (or [0, 2]), while y =
+        # log(MSD) can be any real number.
+        self.parameters = {}
+        for i in range(self.n):
+            self.parameters[f"x{i}"] = Parameter((self.x_first, self.x_last),
+                                                 linearization=Linearize.Bounded(),
+                                                 )
+            self.parameters[f"y{i}"] = Parameter((-np.inf, np.inf),
+                                                 linearization=Linearize.Exponential(),
+                                                 )
+
+        # x0 = x_first and x(n-1) = x_last are always fixed
+        del self.parameters["x0"]
+        del self.parameters[f"x{self.n-1}"]
 
         self.prev_fit = previous_spline_fit_and_result # for (alternative) initialization
 
@@ -133,7 +144,7 @@ class SplineFit(Fit):
 
         See also
         --------
-        SplineFit
+        SplineFit, decompactify_log
         """
         x = np.log(dt) / np.log(self.T)
         if self.ss_order == 0:
@@ -153,6 +164,10 @@ class SplineFit(Fit):
         -------
         log_dt : np.array
             the corresponding ``log(dt [frames])`` values
+
+        See also
+        --------
+        compactify
         """
         if self.ss_order == 0:
             x = np.tan(np.pi/4*x)
@@ -176,8 +191,11 @@ class SplineFit(Fit):
         --------
         params2msdm
         """
-        x = np.array([self.x_first, *params[:(self.n-2)], self.x_last])
-        y = params[(self.n-2):]
+        x = np.array([self.x_first]
+                   + [params[f"x{i}"] for i in range(1, self.n-1)]
+                   + [self.x_last]
+                    )
+        y = [params[f"y{i}"] for i in range(self.n)]
         return interpolate.CubicSpline(x, y, bc_type=self.bc_type)
 
     def params2msdm(self, params):
@@ -222,7 +240,7 @@ class SplineFit(Fit):
 
         See also
         --------
-        Fit.initial_params
+        Fit.initial_params <bayesmsd.fit.Fit.initial_params>
         """
         x_init = np.linspace(self.x_first, self.x_last, self.n)
 
@@ -236,11 +254,11 @@ class SplineFit(Fit):
             # For ss_order == 0 we will use it as boundary condition,
             # for ss_order == 1 this will be the initial MSD
             e_msd = MSD(self.data)
-            t_valid = np.nonzero(~np.isnan(e_msd))[0][1:] # skip msd[0] = 0
+            dt_valid = np.nonzero(~np.isnan(e_msd))[0][1:] # skip msd[0] = 0
             (A, B), _ = optimize.curve_fit(lambda x, A, B : A*x + B,
-                                           self.compactify(t_valid),
-                                           np.log(e_msd[t_valid]),
-                                           p0=(1, np.log(e_msd[1])),
+                                           self.compactify(dt_valid),
+                                           np.log(e_msd[dt_valid]),
+                                           p0=(1, np.log(e_msd[dt_valid[0]])),
                                            bounds=([0, -np.inf], np.inf),
                                           )
                 
@@ -248,7 +266,7 @@ class SplineFit(Fit):
                 # interpolate along 2-point spline
                 ss_var = np.nanmean(np.concatenate([traj.abs()[:][:, 0]**2 for traj in self.data]))
                 csp = interpolate.CubicSpline(np.array([0, 2]),
-                                              np.log([e_msd[1], 2*ss_var]),
+                                              np.log([e_msd[dt_valid[0]], 2*ss_var]),
                                               bc_type = ((1, A), (1, 0.)),
                                              )
                 y_init = csp(x_init)
@@ -256,8 +274,10 @@ class SplineFit(Fit):
                 y_init = A*x_init + B
             else: # pragma: no cover
                 raise ValueError
-            
-        return np.array([*x_init[1:-1], *y_init])
+
+        out = {f"x{i}" : x for i, x in enumerate(x_init[1:-1], start=1)}
+        out.update({f"y{i}" : y for i, y in enumerate(y_init)})
+        return out
 
     def initial_offset(self):
         """
@@ -272,8 +292,6 @@ class SplineFit(Fit):
         else:
             return -self.prev_fit[1]['logL']
         
-    # ensure proper ordering of the x positions
-    # if x's cross, the spline might diverge
     def constraint_dx(self, params):
         """
         Make sure the spline points are properly ordered in x
@@ -297,17 +315,12 @@ class SplineFit(Fit):
         Fit
         """
         min_step = 1e-7 # x is compactified to (0, 1)
-        x = np.array([self.x_first, *params[:(self.n-2)], self.x_last])
+        x = np.array([self.x_first]
+                   + [params[f"x{i}"] for i in range(1, self.n-1)]
+                   + [self.x_last]
+                    )
         return np.min(np.diff(x))/min_step
     
-    # this should (!) be taken care of by the Cpositive constraint
-#     def constraint_dy(self, params):
-#         # Ensure monotonicity in the MSD. This makes sense intuitively, but is it technically a condition?
-#         min_step = 1e-7
-#         y = params[(self.n-2):]
-#         return np.min(np.diff(y))/min_step
-    
-    # make sure that the MSD stays numerically finite
     def constraint_logmsd(self, params):
         """
         Make sure the Spline does not diverge
@@ -334,6 +347,42 @@ class SplineFit(Fit):
         return (full_penalty - np.max(np.abs(csp(x_full))))/start_penalizing
 
 class NPXFit(Fit): # NPX = Noise + Powerlaw + X (i.e. spline)
+    """
+    (N)oise + (P)owerlaw + (X) (i.e. spline) fit
+
+    This scheme can be used to fit powerlaw MSDs with a few extensions: we
+    include localization error and motion blur, and towards long time the MSD
+    might deviate from a powerlaw, which we then fit with a cubic spline (see
+    also `SplineFit`).
+
+    Parameters
+    ----------
+    data : noctiluca.TaggedSet, pandas.DataFrame, list of numpy.ndarray
+        the data to fit. See `Fit`.
+    ss_order : {0, 1}
+        the steady state order to assume
+    n : int >= 0
+        the number of spline points; set ``n = 0`` to fit a pure powerlaw.
+    previous_NPXFit_and_result : tuple (NPXFit, dict)
+        see also `SplineFit`; this can be used to run model selection over
+        ``n``.
+    motion_blur_f : float in [0, 1]
+        fractional exposure time (i.e. exposure time as fraction of the time
+        between frames). The resulting motion blur will be taken into account
+        in the fit.
+
+    Notes
+    -----
+    This fit uses the same compactification scheme as `SplineFit` for the
+    coordinates of the spline nodes.
+
+    Technically, the spline used to extend the MSD has ``n+1`` nodes, since of
+    course there has to be one at the transition from powerlaw to spline.
+
+    The vertical position of the first, and horizontal position of the last
+    spline points are fixed. So the parameters for the spline are ``x0``
+    through ``x(n-1)`` and ``y1`` through ``yn``.
+    """
     def __init__(self, data, ss_order, n=0,
                  previous_NPXFit_and_result=None,
                  motion_blur_f=0,
@@ -344,26 +393,53 @@ class NPXFit(Fit): # NPX = Noise + Powerlaw + X (i.e. spline)
         if n == 0 and ss_order == 0:
             raise ValueError("Incompatible assumptions: pure powerlaw (n=0) and trajectory steady state (ss_order=0)")
         self.n = n
-        
-        # Parameters are (log(noise2), log(Γ), α, x0, ..., x{n-1}, y1, .., yn)
-        # If n == 0 we omit x0. So we always have 2*n spline parameters!
         self.ss_order = ss_order
-        # Take care to not attain the upper bound (2) for the exponent, as this
-        # is where a pure powerlaw stops being positive definite
-        self.bounds = self.d*([(-np.inf, np.inf), (-np.inf, np.inf), (0, 2-1e-10)]
-                              + n*[(0, 2 if ss_order == 0 else 1)] + n*[(-np.inf, np.inf)])
 
+        # Set up parameters
+        # Populate with templates -> expand dimensions -> remove templates
+        # The powerlaw stops being positive definite at α = 2, so stay away from that
+        self.parameters = {
+            'log(σ²)' : Parameter((-np.inf, np.inf),
+                                  linearization=Linearize.Exponential()),
+            'log(Γ)'  : Parameter((-np.inf, np.inf),
+                                  linearization=Linearize.Exponential()),
+            'α'       : Parameter((0, 2-1e-10), # stay away from upper bound
+                                  linearization=Linearize.Bounded()),
+        }
+
+        x_bounds = (0, 2) if self.ss_order == 0 else (0, 1)
+        for i in range(self.n+1):
+            self.parameters[f"x{i}"] = Parameter(x_bounds,
+                                                 linearization=Linearize.Bounded())
+            self.parameters[f"y{i}"] = Parameter((-np.inf, np.inf),
+                                                 linearization=Linearize.Exponential())
+
+        # For the spline, y0 and xn are fixed
+        del self.parameters["y0"]
+        del self.parameters[f"x{self.n}"]
+
+        # Expand dimensions and remove templates
+        # Fix higher dimensions to dim 0, except for localization error
+        param_names = list(self.parameters.keys()) # keys() itself is mutable
+        for name in param_names:
+            for dim in range(self.d):
+                dim_name = f"{name} (dim {dim})"
+                self.parameters[dim_name] = deepcopy(self.parameters[name])
+
+                if name != 'log(σ²)' and dim > 0:
+                    self.parameters[dim_name].fix_to = f"{name} (dim 0)"
+
+            del self.parameters[name]
+
+        # Set up constraints
         self.constraints = [self.constraint_dx,
                             self.constraint_logmsd,
                             self.constraint_Cpositive,
                            ]
-        if self.n == 0: # don't need constraints
+        if self.n == 0: # pure powerlaw; don't need constraints
             self.constraints = []
 
-        self.fix_values = [((2*n+3)*dim + i, lambda x, i=i: x[i])
-                           for dim in range(1, self.d)
-                           for i in range(1, 2*n+3)]
-        
+        # Compactification & splines
         self.logT = np.log(self.T)
         if self.ss_order == 0:
             # Fit in 4/π*arctan(log) space and add point at infinity, i.e. x = 4/π*arctan(log(∞)) = 2
@@ -377,21 +453,82 @@ class NPXFit(Fit): # NPX = Noise + Powerlaw + X (i.e. spline)
             raise ValueError(f"Did not understand ss_order = {ss_order}")
 
         self.prev_fit = previous_NPXFit_and_result # for (alternative) initialization
-        if self.prev_fit and not self.prev_fit[0].d == self.d:
+        if self.prev_fit is not None and not self.prev_fit[0].d == self.d:
             raise ValueError(f"Previous NPXFit has different number of dimensions ({self.prev_fit[0].d}) from the current data set ({self.d}).")
 
     def compactify(self, dt):
+        """
+        Compactification used for the current fit
+
+        Parameters
+        ----------
+        dt : np.array, dtype=float
+            the time lags to calculate compactification for. Should be ``> 0``
+            but might include ``np.inf``.
+
+        See also
+        --------
+        NPXFit, decompactify_log
+        """
         x = np.log(dt) / self.logT
         if self.ss_order == 0:
             x = (4/np.pi)*np.arctan(x)
         return x
 
     def decompactify_log(self, x):
+        """
+        Decompactify spline points (convenience function)
+
+        Parameters
+        ----------
+        x : np.array, dtype=float
+            the compactified x-coordinates
+
+        Returns
+        -------
+        log_dt : np.array
+            the corresponding ``log(dt [frames])`` values
+
+        See also
+        --------
+        compactify
+        """
         if self.ss_order == 0:
             x = np.tan(np.pi/4*x)
         return x * self.logT
 
     def _first_spline_point(self, x0, logG, alpha):
+        """
+        Determine the coordinates of the first spline point
+
+        The first spline point has to be chosen to match the powerlaw before.
+        This function determines y0 and the associated slope (which serves as
+        boundary condition for the spline) from the powerlaw and the cutoff x0.
+        For completeness, x0 is also returned.
+
+        Parameters
+        ----------
+        x0 : float
+            x-coordinate of the first spline point. This is a free parameter.
+        logG, alpha : float
+            the powerlaw at early times that we have to match
+
+        Returns
+        -------
+        x0 : float
+            same as input
+        logt0 : float
+            x0 decompactified (for convenience)
+        y0 : float
+            y0 matching the given powerlaw
+        dcdx0 : float
+            slope of the spline at ``(x0, y0)``, to be used as boundary
+            condition
+
+        See also
+        --------
+        decompactify_log, NPXFit
+        """
         logt0 = self.decompactify_log(x0)
         y0 = alpha*logt0 + logG
 
@@ -406,40 +543,70 @@ class NPXFit(Fit): # NPX = Noise + Powerlaw + X (i.e. spline)
         return x0, logt0, y0, dcdx0
 
     def _params2csp(self, params):
+        """
+        Assemble the cubic spline(s) from given parameters
+
+        Parameters
+        ----------
+        params : dict
+
+        Returns
+        -------
+        list of scipy.interpolate.CubicSpline
+            one for each spatial dimension
+
+        See also
+        --------
+        params2msdm
+        """
         csps = self.d*[None]
         if self.n > 0:
             for dim in range(self.d):
-                params_1d = params[((2*self.n+3)*dim):((2*self.n+3)*(dim+1))]
+                x0, logt0, y0, dcdx0 = self._first_spline_point(
+                                            params[f"x0 (dim {dim})"],
+                                            params[f"log(Γ) (dim {dim})"],
+                                            params[f"α (dim {dim})"],
+                                       )
 
-                x0, logt0, y0, dcdx0 = self._first_spline_point(*params_1d[[3, 1, 2]])
-                x = np.append(params_1d[3:(3+self.n)], self.x_last)
-                y = np.insert(params_1d[(3+self.n):(3+2*self.n)], 0, y0)
+                x = np.array(       [params[  f"x{i} (dim {dim})"] for i in range(self.n)] + [self.x_last]) 
+                y = np.array([y0] + [params[f"y{i+1} (dim {dim})"] for i in range(self.n)])
 
                 csps[dim] = interpolate.CubicSpline(x, y, bc_type=((1, dcdx0), self.upper_bc_type))
 
         return csps
 
     def params2msdm(self, params):
+        """
+        Calculate the current MSD from given parameters
+
+        Parameters
+        ----------
+        params : dict
+
+        Returns
+        -------
+        list of (msd, m)
+
+        See also
+        --------
+        Fit.params2msdm
+        """
         csps = self._params2csp(params)
         msdm = []
         for dim, csp in enumerate(csps):
-            params_1d = params[((2*self.n+3)*dim):((2*self.n+3)*(dim+1))]
-
             with np.errstate(under='ignore'): # if noise == 0
-                noise2, G = np.exp(params_1d[[0, 1]])
-            alpha = params_1d[2]
+                noise2 = np.exp(params[f"log(σ²) (dim {dim})"])
+                G      = np.exp(params[ f"log(Γ) (dim {dim})"])
+            alpha = params[f"α (dim {dim})"]
 
+            # Define "naked" MSD function
             if self.n == 0:
-                @deco.MSDfun
-                @deco.imaging(noise2=noise2, f=self.motion_blur_f, alpha0=alpha)
                 def msd(dt, G=G, alpha=alpha):
                     return G*(dt**alpha)
             else:
-                t0 = np.exp(self.decompactify_log(params_1d[3]))
+                t0 = np.exp(self.decompactify_log(params[f"x0 (dim {dim})"]))
 
-                @deco.MSDfun
-                @deco.imaging(noise2=noise2, f=self.motion_blur_f, alpha0=alpha)
-                def msd(dt, G=G, alpha=alpha, csp=csp):
+                def msd(dt, G=G, alpha=alpha, csp=csp, t0=t0):
                     out = G*(dt**alpha)
                     ind = dt > t0
                     if np.any(ind):
@@ -447,35 +614,55 @@ class NPXFit(Fit): # NPX = Noise + Powerlaw + X (i.e. spline)
                         out[ind] = np.exp(csp(x))
                     return out
 
+            # Apply MSD function decorators
+            msd = deco.imaging(noise2=noise2, f=self.motion_blur_f, alpha0=alpha)(msd)
+            msd = deco.MSDfun(msd)
+
             msdm.append((msd, 0))
         return msdm
             
     def initial_params(self):
-        params = np.empty((3+2*self.n)*self.d, dtype=float)
-        params[:] = np.nan
+        """
+        Provide an initial parameter guess
+
+        These guesses come from a powerlaw curve fit to the empirical MSD. If ``n > 0`` we set ``x0 = 0.5`` and insert evenly spaced spline points above that.
+
+        If ``previous_NPXFit_and_result`` was specified at initialization, we
+        copy the powerlaw and ``x0`` and then place evenly spaced spline points
+        above, reproducing the existing best fit.
+
+        Returns
+        -------
+        params : dict
+        """
+        params = {}
 
         if self.prev_fit:
             fit, res = self.prev_fit
             csps = fit._params2csp(res['params'])
 
             for dim in range(self.d):
-                old_params_1d = res['params'][((2*fit.n+3)*dim):((2*fit.n+3)*(dim+1))]
-                ioff = (2*self.n + 3)*dim
-                params[ioff:(ioff+3)] = old_params_1d[:3]
-                if self.n == 0:
-                    continue
+                for name in ['log(σ²)', 'log(Γ)', 'α']:
+                    params[f"{name} (dim {dim})"] = res['params'][f"{name} (dim {dim})"]
 
-                if fit.n == 0 or old_params_1d[3] >= self.x_last: # note order of conditions: old_params_1d[3] only exists if fit.n > 0
-                    x_init = np.linspace(0.5, self.x_last, self.n+1)
-                    y_init = params[ioff+2]*self.decompactify_log(x_init) + params[ioff+1]
-                else:
-                    x_init = np.linspace(old_params_1d[3], self.x_last, self.n+1)
-                    y_init = csps[dim](x_init)
+                if self.n > 0:
+                    # Note order of conditions: x0 exists only if fit.n > 0
+                    if fit.n == 0 or res['params'][f"x0 (dim {dim})"] >= self.x_last:
+                        logG  = res['params'][f"log(Γ) (dim {dim})"]
+                        alpha = res['params'][     f"α (dim {dim})"]
+
+                        x_init = np.linspace(0.5, self.x_last, self.n+1)
+                        y_init = alpha*self.decompactify_log(x_init) + logG
+                    else:
+                        x0 = res['params'][f"x0 (dim {dim})"]
+                        x_init = np.linspace(x0, self.x_last, self.n+1)
+                        y_init = csps[dim](x_init)
+
+                    params.update({f"x{i} (dim {dim})" : x for i, x in enumerate(x_init[:-1])})
+                    params.update({f"y{i} (dim {dim})" : y for i, y in enumerate(y_init[1:], start=1)})
 
-                params[(ioff+3):(ioff+3+self.n)] = x_init[:-1]
-                params[(ioff+3+self.n):(ioff+3+2*self.n)] = y_init[1:]
         else:
-            # Fit linear (i.e. powerlaw), which is useful in both cases.
+            # Fit linear (i.e. powerlaw), which is useful in both (ss_order) cases.
             # For ss_order == 0 we will use it as boundary condition,
             # for ss_order == 1 this will be the initial MSD
             e_msd = MSD(self.data)/self.d
@@ -488,12 +675,14 @@ class NPXFit(Fit): # NPX = Noise + Powerlaw + X (i.e. spline)
                                               )
 
             for dim in range(self.d):
-                ioff = (2*self.n+3)*dim
-                params[ioff:(ioff+3)] = [np.log(e_msd[1]/2), logG, alpha]
+                params[f"log(σ²) (dim {dim})"] = np.log(e_msd[dt_valid[0]]/2)
+                params[ f"log(Γ) (dim {dim})"] = logG
+                params[      f"α (dim {dim})"] = alpha
 
             if self.n > 0:
                 x0, logt0, y0, dcdx0 = self._first_spline_point(0.5, logG, alpha)
                 x_init = np.linspace(x0, self.x_last, self.n+1)
+
                 if self.ss_order == 0:
                     # interpolate along 2-point spline
                     ss_var = np.nanmean(np.concatenate([np.sum(traj[:]**2, axis=1) for traj in self.data]))/self.d
@@ -508,11 +697,9 @@ class NPXFit(Fit): # NPX = Noise + Powerlaw + X (i.e. spline)
                     raise ValueError
 
                 for dim in range(self.d):
-                    ioff = (2*self.n+3)*dim
-                    params[(ioff+3):(ioff+3+self.n)] = x_init[:-1]
-                    params[(ioff+3+self.n):(ioff+3+2*self.n)] = y_init[1:]
+                    params.update({f"x{i} (dim {dim})" : x for i, x in enumerate(x_init[:-1])})
+                    params.update({f"y{i} (dim {dim})" : y for i, y in enumerate(y_init[1:], start=1)})
 
-        assert ~np.any(np.isnan(params))
         return params
 
     def initial_offset(self):
@@ -527,37 +714,64 @@ class NPXFit(Fit): # NPX = Noise + Powerlaw + X (i.e. spline)
             return -self.prev_fit[1]['logL']
         
     def constraint_dx(self, params):
+        """
+        Make sure the spline points are properly ordered in x
+
+        We impose this constraint mainly to avoid crossing of spline points,
+        which usually leads to the spline diverging. On top of that,
+        conceptually this makes the solution well-defined.
+
+        Parameters
+        ----------
+        params : np.ndarray, dtype=float
+            the current fit parameters
+
+        Returns
+        -------
+        float
+            the constraint score
+
+        See also
+        --------
+        Fit
+        """
         # constraints are not applied if n == 0, so we can safely assume n > 0
         min_step = 1e-7 # x is compactified to (0, 1)
-        x = np.stack([np.append(params[((2*self.n+3)*dim + 3):((2*self.n+3)*dim + 3+self.n)], self.x_last)
-                      for dim in range(self.d)], axis=0)
+        x = np.array([[params[f"x{i} (dim {dim})"] for i in range(self.n)] + [self.x_last]
+                      for dim in range(self.d)])
         return np.min(np.diff(x, axis=-1))/min_step
     
-    # Should (!) be taken care of by the Cpositive constraint
-#     def constraint_dy(self, params):
-#         # Ensure monotonicity in the MSD. This makes sense intuitively, but is it technically a condition?
-#         if self.n == 0:
-#             return np.inf
-#         
-#         min_step = 1e-7
-#         _, _, y0, _ = self._first_spline_point(*params[[3, 1, 2]])
-#         y = np.array([y0, *params[(self.n+3):]])
-#         return np.min(np.diff(y))/min_step
-    
     def constraint_logmsd(self, params):
+        """
+        Make sure the Spline does not diverge
+
+        Parameters
+        ----------
+        params : np.ndarray, dtype=float
+            the current fit parameters
+
+        Returns
+        -------
+        float
+            the constraint score
+
+        See also
+        --------
+        Fit
+        """
         # constraints are not applied if n == 0, so we can safely assume n > 0
         start_penalizing = 200
         full_penalty = 500
 
         csps = self._params2csp(params)
         x_full = self.compactify(np.arange(1, self.T))
-        xs = [x_full[x_full >= params[(2*self.n+3)*dim + 3]] for dim in range(self.d)]
+        xs = [x_full[x_full >= params[f"x0 (dim {dim})"]] for dim in range(self.d)]
 
         if all(len(x) == 0 for x in xs): # pragma: no cover
             return np.inf
 
         logmsd = np.concatenate([csp(x) for csp, x in zip(csps, xs)])
-        return (full_penalty - np.max(np.abs(logmsd)))/start_penalizing
+        return (full_penalty - np.max(np.abs(logmsd)))/(full_penalty - start_penalizing)
 
 class TwoLocusRouseFit(Fit):
     """
@@ -566,59 +780,61 @@ class TwoLocusRouseFit(Fit):
     This class implements a fit for two loci at fixed separation, but on the
     same polymer. A simple model for these dynamics is given by the infinite
     continuous Rouse model, which gives an analytical expression for the MSD
-    (implemented in `!rouse.twoLocusMSD`). Here we provide an implementation to
-    fit this MSD to data.
+    (implemented in `rouse.twoLocusMSD
+    <https://rouse.readthedocs.io/en/latest/rouse.html#rouse.twoLocusMSD>`).
+    Here we provide an implementation to fit this MSD to data.
 
-    The parameters for this MSD are
-
-    - the (squared) localization error ``noise2``
-    - the Rouse scaling prefactor Γ of a single locus on the polymer
-    - the steady state variance J of the distance between the loci. This
-      encodes the chain length between them
-
-    The parameter vector for this fit is thus ``d*[log(noise2), log(Γ),
-    log(J)]``.
-
-    Since all of these are positive quantities with units, it seems natural to
-    perform the fit in log-space. Thus we (effectively) place a 1/x prior on
-    all of them.
-
-    This class allows for separate fits to all the spatial dimensions of the
-    data. In a standard setting this makes sense only for the localization
-    error, so by default we fix Γ and J to be the same for all dimensions.
+    Parameters
+    ----------
+    data : noctiluca.TaggedSet, pandas.DataFrame, list of numpy.ndarray
+        the data to fit. See `Fit`.
+    motion_blur_f : float in [0, 1]
+        fractional exposure time (i.e. exposure time as fraction of the time
+        between frames). The resulting motion blur will be taken into account
+        in the fit.
     """
-    # there was a constraint J > noise2 at some point. I don't remember why we would need this
-    # TODO: check this
-    def __init__(self, data, k=1,
+    def __init__(self, data,
                  motion_blur_f=0,
                 ):
         super().__init__(data)
         self.motion_blur_f = motion_blur_f
         
-        # Parameters are log- (noise2, Γ, J)
         self.ss_order = 0
-        self.bounds = 3*self.d*[(-np.inf, np.inf)]
+        
+        self.parameters = {}
+        for name in ['log(σ²)', 'log(Γ)', 'log(J)']:
+            for dim in range(self.d):
+                dim_name = f"{name} (dim {dim})"
+                self.parameters[f"{name} (dim {dim})"] = Parameter((-np.inf, np.inf),
+                                                                   linearization=Linearize.Exponential())
+
+                if name != 'log(σ²)' and dim > 0:
+                    self.parameters[dim_name].fix_to = f"{name} (dim 0)"
+
         self.constraints = [] # Don't need to check Cpositive, will always be true for Rouse MSDs
-
-        self.fix_values = [(3*dim+i, lambda x, i=i : x[i])
-                           for dim in range(1, self.d)
-                           for i in [1, 2]]
-
-#     def constraint_sigJ(self, params):
-#         logsig2 = params[[3*i   for i in range(self.d)]]
-#         logJs   = params[[3*i+2 for i in range(self.d)]]
-#         return np.sum(logJs - logsig2) * 100
         
     def params2msdm(self, params):
         """
         Give an MSD function (and mean = 0) for given parameters
 
-        Uses ``rouse.twoLocusMSD``
+        Parameters
+        ----------
+        params : dict
+
+        Returns
+        -------
+        list of (msd, m)
+
+        See also
+        --------
+        `rouse.twoLocusMSD <https://rouse.readthedocs.io/en/latest/rouse.html#rouse.twoLocusMSD>`
         """
         msdm = []
         for dim in range(self.d):
             with np.errstate(under='ignore'): # if noise == 0
-                noise2, G, J = np.exp(params[(3*dim):(3*(dim+1))])
+                noise2 = np.exp(params[f"log(σ²) (dim {dim})"])
+                G      = np.exp(params[ f"log(Γ) (dim {dim})"])
+                J      = np.exp(params[ f"log(J) (dim {dim})"])
 
             @deco.MSDfun
             @deco.imaging(noise2=noise2, f=self.motion_blur_f, alpha0=0.5)
@@ -630,113 +846,24 @@ class TwoLocusRouseFit(Fit):
         
     def initial_params(self):
         """
-        Initial parameters from empirical MSD
+        Initial parameters from curve fit to empirical MSD
+
+        Returns
+        -------
+        params : dict
         """
         e_msd = MSD(self.data) / self.d
+        dt_valid = np.nonzero(~np.isnan(e_msd))[0][1:]
+        dt_valid_early = dt_valid[:min(5, len(dt_valid))]
+
         J = np.nanmean(np.concatenate([traj[:]**2 for traj in self.data], axis=0))
-        G = np.nanmean(e_msd[1:5]/np.sqrt(np.arange(1, 5)))
-        noise2 = e_msd[1]/2
+        G = np.nanmean(e_msd[dt_valid_early]/np.sqrt(dt_valid_early))
+        noise2 = e_msd[dt_valid[0]]/2
 
-        return np.log(np.array(self.d*[noise2, G, J]))
-
-class OneLocusRouseFit(Fit):
-    """
-    Fit a Rouse model for one locus on a polymer
-
-    This is very parallel to the implementation of `TwoLocusRouseFit`.
-
-    The parameters for this MSD are
-
-    - the (squared) localization error ``noise2``
-    - the Rouse scaling prefactor Γ
-
-    Since all of these are positive quantities with units, it seems natural to
-    perform the fit in log-space. Thus we (effectively) place a 1/x prior on
-    all of them.
-
-    The parameter vector for this fit is thus ``d*[log(noise2), log(Γ)]``.
-
-    This class allows for separate fits to all the spatial dimensions of the
-    data. In a standard setting this makes sense only for the localization
-    error, so by default we fix Γ to be the same for all dimensions.
-    """
-    # TODO: remove this; if you want to fit a powerlaw with α = 0.5, use NPXFit
-    def __init__(self, data, k=1,
-                 motion_blur_f=0,
-                ):
-        super().__init__(data)
-        self.motion_blur_f = motion_blur_f
-        
-        # Parameters are log- (noise2, Γ)
-        self.ss_order = 1
-        self.bounds = 2*self.d*[(-np.inf, np.inf)]
-        self.constraints = [] # Don't need to check Cpositive, will always be true for Rouse MSDs
-
-        self.fix_values = [(2*dim+i, lambda x, i=i : x[i])
-                           for dim in range(1, self.d)
-                           for i in [1]]
-        
-    def params2msdm(self, params):
-        """
-        Give an MSD function (and drift = 0) for given parameters
-        """
-        msdm = []
+        params = {}
         for dim in range(self.d):
-            with np.errstate(under='ignore'): # if noise == 0
-                noise2, G = np.exp(params[(2*dim):(2*(dim+1))])
+            params[f"log(σ²) (dim {dim})"] = np.log(noise2)
+            params[ f"log(Γ) (dim {dim})"] = np.log(G)
+            params[ f"log(J) (dim {dim})"] = np.log(J)
 
-            @deco.MSDfun
-            @deco.imaging(noise2=noise2, f=self.motion_blur_f, alpha0=0.5)
-            def msd(dt, G=G):
-                return G*np.sqrt(dt)
-
-            msdm.append((msd, 0))
-        return msdm
-        
-    def initial_params(self):
-        """
-        Initial parameters from empirical MSD
-        """
-        e_msd = MSD(self.data) / self.d
-        G = np.nanmean(e_msd[1:5]/np.sqrt(np.arange(1, 5)))
-        noise2 = e_msd[1]/2
-
-        return np.log(np.array(self.d*[noise2, G]))
-
-# class RouseFitDiscrete(Fit):
-#     def __init__(self, data, w):
-#         super().__init__(data)
-#         self.w = w
-#         
-#         # Fit properties
-#         # Parameters are d*[noise2, D, k], where by default we fix Ds and ks to be equal
-#         self.ss_order = 0
-#         self.bounds = 3*self.d*[(1e-10, np.inf)] # List of (lb, ub), to be passed to optimize.minimize
-# 
-# # Note: the following loop does not work, because i is local to the list
-# # comprehension, thus the lambdas will ultimately all return the same value.
-# #         self.fix_values = [(3*dim+i, lambda x: x[i]) for dim in range(1, self.d) for i in [1, 2]]
-#         self.fix_values  = [(3*dim+1, lambda x : x[1]) for dim in range(1, self.d)]
-#         self.fix_values += [(3*dim+2, lambda x : x[2]) for dim in range(1, self.d)]
-# 
-#         self.constraints = [] # Don't need to check Cpositive, will always be true for Rouse MSDs
-#         
-#     def params2acfm(self, params):
-#         acf = np.empty((self.T, self.d), dtype=float)
-#         for dim in range(self.d):
-#             noise2, D, k = params[(3*dim):(3*(dim+1))]
-#             model = rouse.Model(len(self.w), D, k, d=1, setup_dynamics=False)
-#             acf[:, dim] = model.ss_ACF(np.arange(self.T), w=self.w)
-#             acf[0, dim] += noise2
-#         return acf, 0
-#     
-#     def initial_params(self):
-#         e_msd = MSD(self.data) / self.d
-#         J = np.nanmean(np.concatenate([traj[:]**2 for traj in self.data], axis=0))
-#         G = np.nanmean(e_msd[1:5]/np.sqrt(np.arange(1, 5)))
-#         L = np.diff(np.nonzero(self.w)[0])[0]
-#         
-#         D = np.pi*G**2*L/(16*J)
-#         k = np.pi*( G*L / (4*J) )**2
-# 
-#         return np.array(self.d*[e_msd[1]/2, D, k])
+        return params
